#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# --------------------------------------------------------------------------
# Get YouTube Video, Copyright (C) 2006 Arctor Fred, Hirs, Mathias Panzenböck
# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------

# version: 1.5
#
# differnet versions:
#  by arctor (original author):
#     http://www.kde-apps.org/content/show.php?content=37509
#  by Hirs:
#     http://www.kde-apps.org/content/show.php?content=41432
#  by panzi (this version):
#     http://www.kde-apps.org/content/show.php?content=41456
#
# futher credits:
# spanish translation of the menu entries by Víctor Fernández Martínez
# improvements to video encoding by Tejas Dinkar <tejasdinkar AT gmail DOT com>
# tipps about warnings and errors in python 2.3 by Volker Janzen
# (However, I dropped python 2.3 support now.)
# Russian and Ukrainian translations by Ivan Lezhnev Jr.

import urllib, re, os, sys
import gettext, locale
from subprocess import Popen, PIPE
from htmlentitydefs import name2codepoint

import traceback
#from cStringIO    import StringIO
from ConfigParser import ConfigParser
from getopt import gnu_getopt as getopt

def shesc(*s):
	if len(s) == 1:
		return "'%s'" % s[0].replace("'","'\\''")
	else:
		return tuple(("'%s'" % arg.replace("'","'\\''")) for arg in s)

def strconv(f):
	def __strconv(*args,**kwargs):
		return ''.join(f(*args,**kwargs))
	
	return __strconv

@strconv
def htmlesc(s):
	for c in s:
		if c == '<':
			yield "&lt;"
		elif c == '>':
			yield "&gt;"
		elif c == '"':
			yield "&quot;"
		elif c == '&':
			yield "&amp;"
		else:
			yield c

def htmlmap(s):
	if s[0:1] == '#':
		try:
			if s[1:2] == 'x':
				return unichr(int(s[2:],16))
			else:
				return unichr(int(s[1:],10))
		except ValueError:
			pass

	cp = name2codepoint.get(s,None)

	if cp is None:
		return None

	return unichr(cp)

@strconv
def htmlunesc(s):
	i = 0

	while True:
		j = s.find('&',i)

		if j < 0:
			yield s[i:]
			break

		else:
			yield s[i:j]

			i = j
			j = s.find(';',i)

			if j < 0:
				yield '&'
				i += 1

			else:
				tr = htmlmap(s[i+1:j])

				if tr is None:
					yield '&'
					i += 1

				else:
					yield tr
					i = j + 1

def fixendl(s):
	if s[-1:] == '\n':
		return s[:-1]
	else:
		return s

def cmd(*args,**kwargs):
	p = Popen(args,**kwargs)

	if kwargs.get('stderr',None) == PIPE:
		serr = fixendl(p.stderr.read())
	else:
		serr = ''

	if kwargs.get('stdout',None) == PIPE:
		sout = fixendl(p.stdout.read())
	else:
		sout = ''

	return sout, serr, p.wait()

# dialoges:
def showError(caption,msg):
	return cmd('kdialog', '--caption', caption, '--error', msg)

def showSaveFile(caption,file_name):
	return cmd('kdialog', '--caption', caption, '--getsavefilename', file_name, stdout=PIPE)
	
def showWarningYesNoCancel(caption,msg):
	return cmd('kdialog', '--caption', caption, '--warningyesnocancel', msg)

def showPassivePopup(caption,msg,secs=5):
	return cmd('kdialog', '--caption', caption, '--passivepopup', msg, str(secs))

unescDict={'\\':'\\','"':'"',"'":"'",'n':'\n','t':'\t','r':'\r'}

@strconv
def javaScriptUnescape(s):
	n=len(s)
	i=0
	while i < n:
		c = s[i]
		if c == '\\':
			try:
				yield unescDict[s[i+1]]
				i += 1
			except KeyError:
				yield '\\'
		else:
			yield c
		i += 1

def get_filename(save_dir, video_id, format):
	url = "http://www.youtube.com/watch?v=" + video_id
	
	stream = urlopener.open(url)

	try:
		data = stream.read()
	finally:
		stream.close()

	m = re.search("watch_fullscreen\\?(?P<vid_query>.*?)&title=(?P<name>.*)';\n", data)

	if m is None:
		raise Exception(
			_('<p>No video information found on the video\'s Page.<br/>\n'
			'Maybe you have to be logged in to view this video or YouTube has changed it\'s page layout?<br/>\n'
			'If you are already logged in or you don\'t have to be logged in to view this video, '
			'please report this <a href="http://kde-apps.org/content/show.php?content=41456">here</a>.\n'
			'video-page: <a href="%(url)s">%(url)s</a></p>') %
			{'url':url})

	name = javaScriptUnescape(htmlunesc(m.group('name')))
	vid_query = m.group('vid_query')
	
	m = re.search("watchUsername *= *'(?P<by>.*)';\s*\n", data)

	if m:
		name += " by " + javaScriptUnescape(htmlunesc(m.group('by')))
	
	name = name.replace('/', '')

	if format == 'flv':
		ext = 'flv'
	elif format == '5':
		ext = 'flv'
		vid_query += '&fmt=5'
	elif format in ('flv-hq', '6'):
		ext = 'flv'
		vid_query += '&fmt=6'
	elif format in ('3gp', '13'):
		ext = '3gp'
		vid_query += '&fmt=13'
	elif format in ('mp4', '17'):
		ext = 'mp4'
		vid_query += '&fmt=17'
	elif format in ('mp4-hq', '18'):
		ext = 'mp4'
		vid_query += '&fmt=18'
	elif format.isdigit():
		ext = 'flv'
		vid_query += '&fmt=' + format
	else:
		raise Exception(_('Unknown file format: %(format)s') % format)

	return "%s/%s.%s" % (save_dir,name,ext), vid_query

def get_video(vid_query, file_name, action):
	url = "http://www.youtube.com/get_video.php?" + vid_query

	if cmd('kfmclient', 'copy', url, file_name)[2]:
		raise Exception(_('Error saving file: %(file)s') % {'file':file_name})
	
	if action:
		fdir, fname = os.path.split(file_name)
		fname, fext = os.path.splitext(fname)
		vars = {
			'fdir':  shesc(fdir),
			'fname': shesc(fname),
			'fext':  shesc(fext),
			'fpath': shesc(file_name)
		}
		command = action % vars

		p = Popen(command,shell=True,stderr=PIPE)
		code = p.wait()

		if code:
			raise Exception(
				_("<p>Error executing command:<br/>\n"
				"<pre>%(command)s</pre><br/>\n"
				"Command output:<br/>\n"
				"<pre>%(output)s</pre></p>") %
				{'command':htmlesc(command),'output':htmlesc(p.stderr.read())})

def main():
	global urlopener, _

	try:
		conf_file = os.path.join(os.getenv('HOME'),'.get_yt_videorc')

		bin_dir    = os.path.dirname(sys.argv[0])
		prefix     = os.path.dirname(bin_dir)
		locale_dir = os.path.join(prefix,'share','get_yt_video')

		LANG = os.getenv('LANG')
		
		if locale.getlocale() == (None, None) and LANG is not None:
			try:
				lang = gettext.translation('get_yt_video', locale_dir, languages=[LANG])
				lang.install()
			except:
				def _(s): return s

		else:
			gettext.install('get_yt_video', locale_dir)

		# use a FancyURLopener so we can use konquerors cookies :)
		urlopener = urllib.FancyURLopener()

		cookies, serr, code = cmd('dcop', 'kcookiejar', 'kcookiejar', 'findCookies(QString)', 'http://www.youtube.com')

		if code != 0:
			# try to start kcookiejar
			ignore, serr, code = cmd('kcookiejar')

			if code == 0:
				cookies, serr, code = cmd('dcop', 'kcookiejar', 'kcookiejar', 'findCookies(QString)', 'http://www.youtube.com')

		if code == 0 and cookies:
			urlopener.addheader(cookies)

		conf = ConfigParser()

		# parse command line options:
		try:
			opts, args = getopt(sys.argv[1:], "a:f:",
				["action=","format="])
		except Exception, e:
			print >> sys.stderr, '***', str(e)
			sys.exit(2)
		
		if len(args) < 1:
			print >> sys.stderr, '*** no url given'
			sys.exit(2)

		action = ''
		format = 'mp4-hq'

		for o, a in opts:
			if o in ("-a", "--action"):
				action = a
			elif o in ("-f", "--format"):
				format = a.strip()
			else:
				assert False, "unhandled option"
		
		result = set()
		
		for url in args:
			# searching for video id in url:
			m = re.search('youtube\.com/v/(?P<video_id>[-A-Za-z0-9_]+)', url)
			
			if m is None:
				m = re.search('video_id=(?P<video_id>[-A-Za-z0-9_]+)', url)
				
			if m is None:
				m = re.search('v=(?P<video_id>[-A-Za-z0-9_]+)', url)
			
			if m is not None:
				# video id found
				result.add(m.group('video_id'))
	
			else:
				# searching for video ids in content:
				stream = urlopener.open(url)
				
				try:
					data = stream.read()
				finally:
					stream.close()
				
				found = re.findall('youtube\.com/v/([-A-Za-z0-9_]+)', data)
				
				if not found:
					found = re.findall('video_id=([-A-Za-z0-9_]+)', data)
				
				if not found:
					found = re.findall('v=([-A-Za-z0-9_]+)', data)
			
				if not found:
					showError(
						_("Save YouTube Video"),
						_("Can\'t get the youtube-videos ID out of the url or the referred site content.\n"
						  "Maybe you have to be logged in to view this video?"))

				result.update(found)
			
		conf.read([conf_file])

		if conf.has_option('get_yt_video','save_dir'):
			save_dir = conf.get('get_yt_video','save_dir')
		
		else:
			save_dir = os.curdir
		
		if not conf.has_section('get_yt_video'):
			conf.add_section('get_yt_video')
		
		for video_id in result:
			file_name, vid_query = get_filename(save_dir, video_id, format)
				
			while True:
				file_name, serr, code = showSaveFile(_("Save YouTube Video"),file_name)

				if code != 0: # Cancel
					break

				if not os.path.exists(file_name):
					break

				ignore, serr, code = showWarningYesNoCancel(
						_("Save YouTube Video"),
						_('File "%(file)s" already exists.\nDo you want to replace it?') %
						{'file':file_name})
						
				if code == 0: # Yes
					break

				elif code == 1: # No
					continue

				else: # code == 2: Cancel
					break

			if code == 0: # Ok [+ Replace]
				save_dir = os.path.split(file_name)[0]

				try:
					conf.set('get_yt_video','save_dir',save_dir)
					conf.write(open(conf_file,'w'))
				except:
					traceback.print_exc(file=sys.stderr)
			
				try:
					get_video(vid_query, file_name, action)
				
				except Exception, errmsg:
					showError(_('Save YouTube Video'),str(errmsg))

	except Exception, err:
# debug output:
#		sio = StringIO()
#		traceback.print_exc(file=sio)
#		showError(_('Save YouTube Video'),"<pre>" + htmlesc(sio.getvalue()) + "</pre>")
#		sio.close()
		
		traceback.print_exc(file=sys.stderr)
		showError(_('Save YouTube Video'),str(err))

if __name__ == "__main__":
	main()
